# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('..'))


# -- Project information -----------------------------------------------------

project = 'uncertainty-toolbox'
copyright = '2021, uncertainty-toolbox team'
author = 'uncertainty-toolbox team'

# The full version, including alpha/beta/rc tags
release = '0.1.0'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx.ext.imgmath',
    'sphinx_autodoc_typehints',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
import sphinx_rtd_theme

html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
html_theme = 'sphinx_rtd_theme'

# Register the theme as an extension to generate a sitemap.xml
extensions.append("sphinx_rtd_theme")
# logo
html_logo = 'images/logo.svg'
html_favicon = 'assets/square_logo_only.png'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    'style_nav_header_background': 'white',
    'display_version': True,
    'logo_only': False,
    'collapse_navigation': False,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_css_files = [
    "css/custom.css",
]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
html_sidebars = {
  '**': ['logo-text.html', 'globaltoc.html', 'searchbox.html']
  #'**': ['logo-text.html', 'globaltoc.html', 'localtoc.html', 'searchbox.html']
}



"""
Tests for visualizations.
"""

import pytest
import numpy as np
import matplotlib
import matplotlib.pyplot as plt

from uncertainty_toolbox.viz import (
    plot_xy,
    plot_intervals,
    plot_intervals_ordered,
    plot_calibration,
    plot_adversarial_group_calibration,
    plot_sharpness,
    plot_residuals_vs_stds,
    filter_subset,
)


@pytest.fixture
def get_test_set():
    y_pred = np.array([1, 2, 3])
    y_std = np.array([0.1, 0.5, 1])
    y_true = np.array([1.5, 3, 2])
    x = np.array([4, 5, 6.5])
    return y_pred, y_std, y_true, x


@pytest.fixture
def get_fig_ax():
    fig, ax = plt.subplots(1, 1, figsize=(5, 5))
    return fig, ax


def test_plot_xy_returns(get_test_set, get_fig_ax):
    """Test if plot_xy returns correct type."""
    y_pred, y_std, y_true, x = get_test_set
    fig, ax = get_fig_ax
    ax = plot_xy(y_pred, y_std, y_true, x, ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)


def test_plot_intervals_returns(get_test_set, get_fig_ax):
    """Test if plot_intervals returns correct type."""
    y_pred, y_std, y_true, _ = get_test_set
    fig, ax = get_fig_ax
    ax = plot_intervals(y_pred, y_std, y_true, ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)


def test_plot_intervals_ordered_returns(get_test_set, get_fig_ax):
    """Test if plot_intervals_ordered returns correct type."""
    y_pred, y_std, y_true, _ = get_test_set
    fig, ax = get_fig_ax
    ax = plot_intervals_ordered(y_pred, y_std, y_true, ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)


def test_plot_calibration_returns(get_test_set, get_fig_ax):
    """Test if plot_calibration returns correct type."""
    y_pred, y_std, y_true, _ = get_test_set
    fig, ax = get_fig_ax
    ax = plot_calibration(y_pred, y_std, y_true, ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)


def test_plot_adversarial_group_calibration_returns(get_test_set, get_fig_ax):
    """Test if plot_adversarial_group_calibration returns correct type."""
    y_pred, y_std, y_true, _ = get_test_set
    fig, ax = get_fig_ax
    ax = plot_adversarial_group_calibration(y_pred, y_std, y_true, ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)


def test_plot_sharpness_returns(get_test_set, get_fig_ax):
    """Test if plot_adversarial_group_calibration returns correct type."""
    _, y_std, _, _ = get_test_set
    fig, ax = get_fig_ax
    ax = plot_sharpness(y_std, ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)


def test_plot_sharpness_nsubset_returns(get_test_set, get_fig_ax):
    """Test if plot_adversarial_group_calibration returns correct type."""
    _, y_std, _, _ = get_test_set
    fig, ax = get_fig_ax
    _test_n_subset = 2
    ax = plot_sharpness(y_std, n_subset=_test_n_subset, ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)


def test_plot_residuals_vs_stds_returns(get_test_set, get_fig_ax):
    """Test if plot_residuals_vs_stds returns correct type."""
    y_pred, y_std, y_true, _ = get_test_set
    fig, ax = get_fig_ax
    ax = plot_residuals_vs_stds(y_pred, y_std, y_true, ax=ax)
    assert isinstance(ax, matplotlib.axes.Axes)


def test_filter_subset(get_test_set):
    """Test if filter_subset returns correct number of subset elements."""
    y_pred, y_std, y_true, _ = get_test_set
    _test_n_subset = 2
    [y_pred, y_std, y_true] = filter_subset([y_pred, y_std, y_true], _test_n_subset)
    assert len(y_pred) == _test_n_subset
    assert len(y_std) == _test_n_subset
    assert len(y_true) == _test_n_subset



"""
Util functions for the toolbox.
"""
from typing import Any, NoReturn, Tuple, Union

import numpy as np

Numeric = Union[int, float, np.ndarray]


def assert_is_flat_same_shape(*args: Any) -> Union[bool, NoReturn]:
    """Check if inputs are all same-length 1d numpy.ndarray.

    Args:
        args: the numpy arrays to check.

    Returns:
        True if all arrays are flat and the same shape, or else raises assertion error.
    """
    assert len(args) > 0
    assert isinstance(args[0], np.ndarray), "All inputs must be of type numpy.ndarray"
    first_shape = args[0].shape
    for arr in args:
        assert isinstance(arr, np.ndarray), "All inputs must be of type numpy.ndarray"
        assert len(arr.shape) == 1, "All inputs must be 1d numpy.ndarray"
        assert arr.shape == first_shape, "All inputs must have the same length"

    return True


def assert_is_positive(*args: Any) -> Union[bool, NoReturn]:
    """Assert that all numpy arrays are positive.

    Args:
        args: the numpy arrays to check.

    Returns:
        True if all elements in all arrays are positive values, or else raises assertion error.
    """
    assert len(args) > 0
    for arr in args:
        assert isinstance(arr, np.ndarray), "All inputs must be of type numpy.ndarray"
        assert np.all(arr > 0.0)

    return True


def trapezoid_area(
    xl: np.ndarray,
    al: np.ndarray,
    bl: np.ndarray,
    xr: np.ndarray,
    ar: np.ndarray,
    br: np.ndarray,
    absolute: bool = True,
) -> Numeric:
    """
    Calculate the area of a vertical-sided trapezoid, formed connecting the following points:
        (xl, al) - (xl, bl) - (xr, br) - (xr, ar) - (xl, al)

    This function considers the case that the edges of the trapezoid might cross,
    and explicitly accounts for this.

    Args:
        xl: The x coordinate of the left-hand points of the trapezoid
        al: The y coordinate of the first left-hand point of the trapezoid
        bl: The y coordinate of the second left-hand point of the trapezoid
        xr: The x coordinate of the right-hand points of the trapezoid
        ar: The y coordinate of the first right-hand point of the trapezoid
        br: The y coordinate of the second right-hand point of the trapezoid
        absolute: Whether to calculate the absolute area, or allow a negative area (e.g. if a and b are swapped)

    Returns: The area of the given trapezoid.

    """

    # Differences
    dl = bl - al
    dr = br - ar

    # The ordering is the same for both iff they do not cross.
    cross = dl * dr < 0

    # Treat the degenerate case as a trapezoid
    cross = cross * (1 - ((dl == 0) * (dr == 0)))

    # trapezoid for non-crossing lines
    area_trapezoid = (xr - xl) * 0.5 * ((bl - al) + (br - ar))
    if absolute:
        area_trapezoid = np.abs(area_trapezoid)

    # Hourglass for crossing lines.
    # NaNs should only appear in the degenerate and parallel cases.
    # Those NaNs won't get through the final multiplication so it's ok.
    with np.errstate(divide="ignore", invalid="ignore"):
        x_intersect = intersection((xl, bl), (xr, br), (xl, al), (xr, ar))[0]
    tl_area = 0.5 * (bl - al) * (x_intersect - xl)
    tr_area = 0.5 * (br - ar) * (xr - x_intersect)
    if absolute:
        area_hourglass = np.abs(tl_area) + np.abs(tr_area)
    else:
        area_hourglass = tl_area + tr_area

    # The nan_to_num function allows us to do 0 * nan = 0
    return (1 - cross) * area_trapezoid + cross * np.nan_to_num(area_hourglass)


def intersection(
    p1: Tuple[Numeric, Numeric],
    p2: Tuple[Numeric, Numeric],
    p3: Tuple[Numeric, Numeric],
    p4: Tuple[Numeric, Numeric],
) -> Tuple[Numeric, Numeric]:
    """
    Calculate the intersection of two lines between four points, as defined in
    https://en.wikipedia.org/wiki/Line%E2%80%93line_intersection.

    This is an array option and works can be used to calculate the intersections of
    entire arrays of points at the same time.

    Args:
        p1: The point (x1, y1), first point of Line 1
        p2: The point (x2, y2), second point of Line 1
        p3: The point (x3, y3), first point of Line 2
        p4: The point (x4, y4), second point of Line 2

    Returns: The point of intersection of the two lines, or (np.nan, np.nan) if the lines are parallel

    """

    x1, y1 = p1
    x2, y2 = p2
    x3, y3 = p3
    x4, y4 = p4

    D = (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4)

    x = ((x1 * y2 - y1 * x2) * (x3 - x4) - (x1 - x2) * (x3 * y4 - y3 * x4)) / D
    y = ((x1 * y2 - y1 * x2) * (y3 - y4) - (y1 - y2) * (x3 * y4 - y3 * x4)) / D

    return x, y



"""Uncertainty Toolbox: a python toolbox for predictive uncertainty quantification, calibration, metrics, and visualization."""

__version__ = "0.1.1rc2"

from .data import (
    synthetic_arange_random,
    synthetic_sine_heteroscedastic,
)

from .metrics import (
    get_all_metrics,
    get_all_accuracy_metrics,
    get_all_average_calibration,
    get_all_adversarial_group_calibration,
    get_all_sharpness_metrics,
    get_all_scoring_rule_metrics,
)

from .metrics_accuracy import prediction_error_metrics

from .metrics_calibration import (
    root_mean_squared_calibration_error,
    mean_absolute_calibration_error,
    miscalibration_area,
    adversarial_group_calibration,
    sharpness,
    get_proportion_lists_vectorized,
    get_proportion_lists,
    get_proportion_in_interval,
    get_prediction_interval,
)

from .metrics_scoring_rule import (
    nll_gaussian,
    crps_gaussian,
    check_score,
    interval_score,
)

from .recalibration import (
    iso_recal,
    optimize_recalibration_ratio,
)

from .viz import (
    plot_xy,
    plot_intervals,
    plot_intervals_ordered,
    plot_calibration,
    plot_adversarial_group_calibration,
    plot_sharpness,
    plot_residuals_vs_stds,
)



"""
Test for util functions.
"""

import pytest
import numpy as np

from uncertainty_toolbox.utils import (
    assert_is_flat_same_shape,
    assert_is_positive,
    trapezoid_area,
)


def test_is_flat_same_shape_wrong_type():
    wrong = [1, 2, 3]
    with pytest.raises(AssertionError):
        assert_is_flat_same_shape(wrong)


def test_is_flat_same_shape_wrong_shape():
    wrong = np.arange(9).reshape(3, 3)
    with pytest.raises(AssertionError):
        assert_is_flat_same_shape(wrong)


def test_is_flat_same_shape_wrong_type_and_shape():
    first = np.arange(3)
    wrong = np.arange(3).reshape(1, 3)
    with pytest.raises(AssertionError):
        assert_is_flat_same_shape(first, wrong)


def test_is_flat_same_shape_not_all_same():
    first = np.arange(3)
    wrong = np.arange(5)
    with pytest.raises(AssertionError):
        assert_is_flat_same_shape(first, wrong)


def test_is_flat_same_shape_correct_many_inputs():
    inputs = [np.arange(5) for _ in range(5)]
    assert_is_flat_same_shape(*inputs)


def test_trapezoid_area_convex():
    # convex trapezoid
    _X = np.array([0, 1, 2, 3])
    _A = np.array([0, 0, 0, 0])
    _B = np.array([0, 1, 2, 0])
    xl, al, bl = _X[:-1], _A[:-1], _B[:-1]
    xr, ar, br = _X[1:], _A[1:], _B[1:]
    area_arr = trapezoid_area(xl, al, bl, xr, ar, br)
    assert area_arr.shape == (len(_X) - 1,)
    assert np.sum(area_arr) == pytest.approx(3.0, abs=1e-10)


def test_trapezoid_area_crossed():
    # crossed trapezoid
    _X = np.array([0, 1, 2, 3])
    _A = np.array([0, 0, 0, 0])
    _B = np.array([0, 1, -1, 0])
    xl, al, bl = _X[:-1], _A[:-1], _B[:-1]
    xr, ar, br = _X[1:], _A[1:], _B[1:]
    area_arr = trapezoid_area(xl, al, bl, xr, ar, br)
    assert area_arr.shape == (len(_X) - 1,)
    assert np.sum(area_arr) == pytest.approx(1.5, abs=1e-10)


def test_trapezoid_area_sine_crossed():
    # sine wave, 1 period, crossed trapezoid
    _NUM_DISCRETIZATION = 1000
    _X = np.linspace(0, 2 * np.pi, _NUM_DISCRETIZATION)
    sine_x = np.sin(_X)
    _A = np.zeros_like(_X)
    _B = sine_x
    xl, al, bl = _X[:-1], _A[:-1], _B[:-1]
    xr, ar, br = _X[1:], _A[1:], _B[1:]
    area_arr = trapezoid_area(xl, al, bl, xr, ar, br)
    assert area_arr.shape == (len(_X) - 1,)
    assert np.sum(area_arr) == pytest.approx(4, abs=1e-4)


def test_trapezoid_area_sine_crossed_rotated():
    # sine wave, 5 periods, crossed trapezoid, 45 deg rotated base
    _NUM_DISCRETIZATION = 10000
    _ROT_DEG = -np.pi / 4
    _X = np.linspace(0, 10 * np.pi, _NUM_DISCRETIZATION)
    sine_x = np.sin(_X)
    _A = np.zeros_like(_X)
    _B = sine_x
    line1 = np.stack([_X, _A]).T
    line2 = np.stack([_X, _B]).T
    rot_mat = np.array(
        [[np.cos(_ROT_DEG), -np.sin(_ROT_DEG)], [np.sin(_ROT_DEG), np.cos(_ROT_DEG)]]
    )
    rot_line1 = line1 @ rot_mat
    rot_line2 = line2 @ rot_mat
    x = rot_line1[:, 0]
    a = rot_line1[:, 1]
    b = rot_line2[:, 1]
    xl, al, bl = x[:-1], a[:-1], b[:-1]
    xr, ar, br = x[1:], a[1:], b[1:]
    area_arr = trapezoid_area(xl, al, bl, xr, ar, br)
    assert area_arr.shape == (len(_X) - 1,)
    assert np.sum(area_arr) == pytest.approx(10, abs=1e-4)


def test_is_flat_same_shape_correct_single_input():
    input = np.arange(5)
    assert_is_flat_same_shape(input)


def test_is_flat_same_shape_correct_many_empty_inputs():
    inputs = [np.arange(0) for _ in range(5)]
    assert_is_flat_same_shape(*inputs)


def test_is_flat_same_shape_correct_single_empty_input():
    input = np.arange(0)
    assert_is_flat_same_shape(input)


def test_is_flat_same_shape_empty_call():
    with pytest.raises(AssertionError):
        assert_is_flat_same_shape()


def test_assert_is_positive_wrong_type():
    wrong = [1, 2, 3]
    with pytest.raises(AssertionError):
        assert_is_positive(wrong)


def test_assert_is_positive_with_zero_as_input():
    wrong = np.arange(9)
    with pytest.raises(AssertionError):
        assert_is_positive(wrong)


def test_assert_is_positive_with_negative_inputs():
    wrong = np.arange(-9, 9, 2)
    with pytest.raises(AssertionError):
        assert_is_positive(wrong)


def test_assert_is_positive_correct_many_inputs():
    inputs = [np.arange(1, 9) for _ in range(5)]
    assert_is_positive(*inputs)


def test_assert_is_positive_correct_single_input():
    input = np.arange(1, 9)
    assert_is_positive(input)


def test_assert_is_positive_correct_2D_input():
    input = np.arange(1, 10).reshape(3, 3)
    assert_is_positive(input)


def test_assert_is_positive_empty_call():
    with pytest.raises(AssertionError):
        assert_is_positive()



"""
Visualizations for predictive uncertainties and metrics.
"""
from typing import Union, Tuple, List, Any, NoReturn
import pathlib

import numpy as np
from scipy import stats
import matplotlib
import matplotlib.pyplot as plt

from uncertainty_toolbox.metrics_calibration import (
    get_proportion_lists,
    get_proportion_lists_vectorized,
    adversarial_group_calibration,
    miscalibration_area,
    miscalibration_area_from_proportions,
)


def plot_xy(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    x: np.ndarray,
    n_subset: Union[int, None] = None,
    ylims: Union[Tuple[float, float], None] = None,
    xlims: Union[Tuple[float, float], None] = None,
    num_stds_confidence_bound: int = 2,
    leg_loc: Union[int, str] = 3,
    ax: Union[matplotlib.axes.Axes, None] = None,
) -> matplotlib.axes.Axes:
    """Plot one-dimensional inputs with associated predicted values, predictive
    uncertainties, and true values.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        x: 1D array of input values for the held out dataset.
        n_subset: Number of points to plot after filtering.
        ylims: a tuple of y axis plotting bounds, given as (lower, upper).
        xlims: a tuple of x axis plotting bounds, given as (lower, upper).
        num_stds_confidence_bound: width of confidence band, in terms of number of
            standard deviations.
        leg_loc: location of legend as a str or legend code int.
        ax: matplotlib.axes.Axes object.

    Returns:
        matplotlib.axes.Axes object with plot added.
    """
    # Create ax if it doesn't exist
    if ax is None:
        fig, ax = plt.subplots(figsize=(5, 5))

    # Order points in order of increasing x
    order = np.argsort(x)
    y_pred, y_std, y_true, x = (
        y_pred[order],
        y_std[order],
        y_true[order],
        x[order],
    )

    # Optionally select a subset
    if n_subset is not None:
        [y_pred, y_std, y_true, x] = filter_subset([y_pred, y_std, y_true, x], n_subset)

    intervals = num_stds_confidence_bound * y_std

    h1 = ax.plot(x, y_true, ".", mec="#ff7f0e", mfc="None")
    h2 = ax.plot(x, y_pred, "-", c="#1f77b4", linewidth=2)
    h3 = ax.fill_between(
        x,
        y_pred - intervals,
        y_pred + intervals,
        color="lightsteelblue",
        alpha=0.4,
    )
    ax.legend(
        [h1[0], h2[0], h3],
        ["Observations", "Predictions", "$95\%$ Interval"],
        loc=leg_loc,
    )

    # Format plot
    if ylims is not None:
        ax.set_ylim(ylims)

    if xlims is not None:
        ax.set_xlim(xlims)

    ax.set_xlabel("$x$")
    ax.set_ylabel("$y$")
    ax.set_title("Confidence Band")
    ax.set_aspect(1.0 / ax.get_data_ratio(), adjustable="box")

    return ax


def plot_intervals(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    n_subset: Union[int, None] = None,
    ylims: Union[Tuple[float, float], None] = None,
    num_stds_confidence_bound: int = 2,
    ax: Union[matplotlib.axes.Axes, None] = None,
) -> matplotlib.axes.Axes:
    """Plot predictions and predictive intervals versus true values.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        n_subset: Number of points to plot after filtering.
        ylims: a tuple of y axis plotting bounds, given as (lower, upper).
        num_stds_confidence_bound: width of intervals, in terms of number of standard
            deviations.
        ax: matplotlib.axes.Axes object.

    Returns:
        matplotlib.axes.Axes object with plot added.
    """
    # Create ax if it doesn't exist
    if ax is None:
        fig, ax = plt.subplots(figsize=(5, 5))

    # Optionally select a subset
    if n_subset is not None:
        [y_pred, y_std, y_true] = filter_subset([y_pred, y_std, y_true], n_subset)

    # Compute intervals
    intervals = num_stds_confidence_bound * y_std

    # Plot
    ax.errorbar(
        y_true,
        y_pred,
        intervals,
        fmt="o",
        ls="none",
        linewidth=1.5,
        c="#1f77b4",
        alpha=0.5,
    )
    h1 = ax.plot(y_true, y_pred, "o", c="#1f77b4")

    # Determine lims
    if ylims is None:
        intervals_lower_upper = [y_pred - intervals, y_pred + intervals]
        lims_ext = [
            int(np.floor(np.min(intervals_lower_upper[0]))),
            int(np.ceil(np.max(intervals_lower_upper[1]))),
        ]
    else:
        lims_ext = ylims

    # plot 45-degree line
    h2 = ax.plot(lims_ext, lims_ext, "--", linewidth=1.5, c="#ff7f0e")

    # Legend
    ax.legend([h1[0], h2[0]], ["Predictions", "$f(x) = x$"], loc=4)

    # Format plot
    ax.set_xlim(lims_ext)
    ax.set_ylim(lims_ext)
    ax.set_xlabel("Observed Values")
    ax.set_ylabel("Predicted Values and Intervals")
    ax.set_title("Prediction Intervals")
    ax.set_aspect(1.0 / ax.get_data_ratio(), adjustable="box")

    return ax


def plot_intervals_ordered(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    n_subset: Union[int, None] = None,
    ylims: Union[Tuple[float, float], None] = None,
    num_stds_confidence_bound: int = 2,
    ax: Union[matplotlib.axes.Axes, None] = None,
) -> matplotlib.axes.Axes:
    """Plot predictions and predictive intervals versus true values, with points ordered
    by true value along x-axis.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        n_subset: Number of points to plot after filtering.
        ylims: a tuple of y axis plotting bounds, given as (lower, upper).
        num_stds_confidence_bound: width of intervals, in terms of number of standard
            deviations.
        ax: matplotlib.axes.Axes object.

    Returns:
        matplotlib.axes.Axes object with plot added.
    """
    # Create ax if it doesn't exist
    if ax is None:
        fig, ax = plt.subplots(figsize=(5, 5))

    # Optionally select a subset
    if n_subset is not None:
        [y_pred, y_std, y_true] = filter_subset([y_pred, y_std, y_true], n_subset)

    order = np.argsort(y_true.flatten())
    y_pred, y_std, y_true = y_pred[order], y_std[order], y_true[order]
    xs = np.arange(len(order))
    intervals = num_stds_confidence_bound * y_std

    # Plot
    ax.errorbar(
        xs,
        y_pred,
        intervals,
        fmt="o",
        ls="none",
        linewidth=1.5,
        c="#1f77b4",
        alpha=0.5,
    )
    h1 = ax.plot(xs, y_pred, "o", c="#1f77b4")
    h2 = ax.plot(xs, y_true, "--", linewidth=2.0, c="#ff7f0e")

    # Legend
    ax.legend([h1[0], h2[0]], ["Predicted Values", "Observed Values"], loc=4)

    # Determine lims
    if ylims is None:
        intervals_lower_upper = [y_pred - intervals, y_pred + intervals]
        lims_ext = [
            int(np.floor(np.min(intervals_lower_upper[0]))),
            int(np.ceil(np.max(intervals_lower_upper[1]))),
        ]
    else:
        lims_ext = ylims

    # Format plot
    ax.set_ylim(lims_ext)
    ax.set_xlabel("Index (Ordered by Observed Value)")
    ax.set_ylabel("Predicted Values and Intervals")
    ax.set_title("Ordered Prediction Intervals")
    ax.set_aspect(1.0 / ax.get_data_ratio(), adjustable="box")

    return ax


def plot_calibration(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    n_subset: Union[int, None] = None,
    curve_label: Union[str, None] = None,
    vectorized: bool = True,
    exp_props: Union[np.ndarray, None] = None,
    obs_props: Union[np.ndarray, None] = None,
    ax: Union[matplotlib.axes.Axes, None] = None,
    prop_type: str = "interval",
) -> matplotlib.axes.Axes:
    """Plot the observed proportion vs prediction proportion of outputs falling into a
    range of intervals, and display miscalibration area.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        n_subset: Number of points to plot after filtering.
        curve_label: legend label str for calibration curve.
        vectorized: plot using get_proportion_lists_vectorized.
        exp_props: plot using the given expected proportions.
        obs_props: plot using the given observed proportions.
        ax: matplotlib.axes.Axes object.
        prop_type: "interval" to measure observed proportions for centered prediction intervals,
                   and "quantile" for observed proportions below a predicted quantile.
                   Ignored if exp_props and obs_props are provided as inputs.

    Returns:
        matplotlib.axes.Axes object with plot added.
    """
    # Create ax if it doesn't exist
    if ax is None:
        fig, ax = plt.subplots(figsize=(5, 5))

    # Optionally select a subset
    if n_subset is not None:
        [y_pred, y_std, y_true] = filter_subset([y_pred, y_std, y_true], n_subset)

    if (exp_props is None) or (obs_props is None):
        # Compute exp_proportions and obs_proportions
        if vectorized:
            (
                exp_proportions,
                obs_proportions,
            ) = get_proportion_lists_vectorized(
                y_pred, y_std, y_true, prop_type=prop_type
            )
        else:
            (exp_proportions, obs_proportions) = get_proportion_lists(
                y_pred, y_std, y_true, prop_type=prop_type
            )
    else:
        # If expected and observed proportions are given
        exp_proportions = np.array(exp_props).flatten()
        obs_proportions = np.array(obs_props).flatten()
        if exp_proportions.shape != obs_proportions.shape:
            raise RuntimeError("exp_props and obs_props shape mismatch")

    # Set label
    if curve_label is None:
        curve_label = "Predictor"

    # Plot
    ax.plot([0, 1], [0, 1], "--", label="Ideal", c="#ff7f0e")
    ax.plot(exp_proportions, obs_proportions, label=curve_label, c="#1f77b4")
    ax.fill_between(exp_proportions, exp_proportions, obs_proportions, alpha=0.2)

    # Format plot
    ax.set_xlabel("Predicted Proportion in Interval")
    ax.set_ylabel("Observed Proportion in Interval")
    ax.axis("square")

    buff = 0.01
    ax.set_xlim([0 - buff, 1 + buff])
    ax.set_ylim([0 - buff, 1 + buff])

    ax.set_title("Average Calibration")

    # Compute miscalibration area
    miscalibration_area = miscalibration_area_from_proportions(
        exp_proportions=exp_proportions, obs_proportions=obs_proportions
    )

    # Annotate plot with the miscalibration area
    ax.text(
        x=0.95,
        y=0.05,
        s="Miscalibration area = %.2f" % miscalibration_area,
        verticalalignment="bottom",
        horizontalalignment="right",
        fontsize="small",
    )

    return ax


def plot_adversarial_group_calibration(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    n_subset: Union[int, None] = None,
    cali_type: str = "mean_abs",
    curve_label: Union[str, None] = None,
    group_size: Union[np.ndarray, None] = None,
    score_mean: Union[np.ndarray, None] = None,
    score_stderr: Union[np.ndarray, None] = None,
    ax: Union[matplotlib.axes.Axes, None] = None,
) -> matplotlib.axes.Axes:
    """Plot adversarial group calibration plots by varying group size from 0% to 100% of
    dataset size and recording the worst calibration occurred for each group size.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        n_subset: Number of points to plot after filtering.
        cali_type: Calibration type str.
        curve_label: legend label str for calibration curve.
        group_size: 1D array of group size ratios in [0, 1].
        score_mean: 1D array of metric means for group size ratios in group_size.
        score_stderr: 1D array of metric standard devations for group size ratios in group_size.
        ax: matplotlib.axes.Axes object.

    Returns:
        matplotlib.axes.Axes object with plot added.
    """
    # Create ax if it doesn't exist
    if ax is None:
        fig, ax = plt.subplots(figsize=(7, 5))

    # Optionally select a subset
    if n_subset is not None:
        [y_pred, y_std, y_true] = filter_subset([y_pred, y_std, y_true], n_subset)

    # Compute group_size, score_mean, score_stderr
    if (group_size is None) or (score_mean is None):
        # Compute adversarial group calibration
        adv_group_cali_namespace = adversarial_group_calibration(
            y_pred, y_std, y_true, cali_type=cali_type
        )
        group_size = adv_group_cali_namespace.group_size
        score_mean = adv_group_cali_namespace.score_mean
        score_stderr = adv_group_cali_namespace.score_stderr
    else:
        # If expected and observed proportions are give
        group_size = np.array(group_size).flatten()
        score_mean = np.array(score_mean).flatten()
        score_stderr = np.array(score_stderr).flatten()
        if (group_size.shape != score_mean.shape) or (
            group_size.shape != score_stderr.shape
        ):
            raise RuntimeError(
                "Input arrays for adversarial group calibration shape mismatch"
            )

    # Set label
    if curve_label is None:
        curve_label = "Predictor"

    # Plot
    ax.plot(group_size, score_mean, "-o", label=curve_label, c="#1f77b4")
    ax.fill_between(
        group_size,
        score_mean - score_stderr,
        score_mean + score_stderr,
        alpha=0.2,
    )

    # Format plot
    buff = 0.02
    ax.set_xlim([0 - buff, 1 + buff])
    ax.set_ylim([0 - buff, 0.5 + buff])
    ax.set_xlabel("Group size")
    ax.set_ylabel("Calibration Error of Worst Group")
    ax.set_title("Adversarial Group Calibration")

    return ax


def plot_sharpness(
    y_std: np.ndarray,
    n_subset: Union[int, None] = None,
    ax: Union[matplotlib.axes.Axes, None] = None,
) -> matplotlib.axes.Axes:
    """Plot sharpness of the predictive uncertainties.

    Args:
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        n_subset: Number of points to plot after filtering.
        ax: matplotlib.axes.Axes object.

    Returns:
        matplotlib.axes.Axes object with plot added.
    """
    # Create ax if it doesn't exist
    if ax is None:
        fig, ax = plt.subplots(figsize=(5, 5))

    # Optionally select a subset
    if n_subset is not None:
        y_std = filter_subset([y_std], n_subset)[0]

    # Plot sharpness curve
    ax.hist(y_std, edgecolor="#1f77b4", color="#a5c8e1", density=True)

    # Format plot
    xlim = (y_std.min(), y_std.max())
    ax.set_xlim(xlim)
    ax.set_xlabel("Predicted Standard Deviation")
    ax.set_ylabel("Normalized Frequency")
    ax.set_title("Sharpness")
    ax.set_yticklabels([])
    ax.set_yticks([])

    # Calculate and report sharpness
    sharpness = np.sqrt(np.mean(y_std**2))
    ax.axvline(x=sharpness, label="sharpness", color="k", linewidth=2, ls="--")

    if sharpness < (xlim[0] + xlim[1]) / 2:
        text = "\n  Sharpness = %.2f" % sharpness
        h_align = "left"
    else:
        text = "\nSharpness = %.2f  " % sharpness
        h_align = "right"

    ax.text(
        x=sharpness,
        y=ax.get_ylim()[1],
        s=text,
        verticalalignment="top",
        horizontalalignment=h_align,
        fontsize="small",
    )

    return ax


def plot_residuals_vs_stds(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    n_subset: Union[int, None] = None,
    ax: Union[matplotlib.axes.Axes, None] = None,
) -> matplotlib.axes.Axes:
    """Plot absolute value of the prediction residuals versus standard deviations of the
    predictive uncertainties.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of the predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        n_subset: Number of points to plot after filtering.
        ax: matplotlib.axes.Axes object.

    Returns:
        matplotlib.axes.Axes object with plot added.
    """
    # Create ax if it doesn't exist
    if ax is None:
        fig, ax = plt.subplots(figsize=(5, 5))

    # Optionally select a subset
    if n_subset is not None:
        [y_pred, y_std, y_true] = filter_subset([y_pred, y_std, y_true], n_subset)

    # Compute residuals
    residuals = y_true - y_pred

    # Put stds on same scale as residuals
    residuals_sum = np.sum(np.abs(residuals))
    y_std_scaled = (y_std / np.sum(y_std)) * residuals_sum

    # Plot residuals vs standard devs
    h1 = ax.plot(y_std_scaled, np.abs(residuals), "o", c="#1f77b4")

    # Plot 45-degree line
    xlims = ax.get_xlim()
    ylims = ax.get_ylim()
    lims = [np.min([xlims[0], ylims[0]]), np.max([xlims[1], ylims[1]])]
    h2 = ax.plot(lims, lims, "--", c="#ff7f0e")

    # Legend
    ax.legend([h1[0], h2[0]], ["Predictions", "$f(x) = x$"], loc=4)

    # Format plot
    ax.set_xlabel("Standard Deviations (Scaled)")
    ax.set_ylabel("Residuals (Absolute Value)")
    ax.set_title("Residuals vs. Predictive Standard Deviations")
    ax.set_xlim(lims)
    ax.set_ylim(lims)
    ax.axis("square")

    return ax


def filter_subset(input_list: List[List[Any]], n_subset: int) -> List[List[Any]]:
    """Keep only n_subset random indices from all lists given in input_list.

    Args:
        input_list: list of lists.
        n_subset: Number of points to plot after filtering.

    Returns:
        List of all input lists with sizes reduced to n_subset.
    """
    assert type(n_subset) is int
    n_total = len(input_list[0])
    idx = np.random.choice(range(n_total), n_subset, replace=False)
    idx = np.sort(idx)
    output_list = []
    for inp in input_list:
        outp = inp[idx]
        output_list.append(outp)
    return output_list


def set_style(style_str: str = "default") -> NoReturn:
    """Set the matplotlib plotting style.

    Args:
        style_str: string for style file.
    """
    if style_str == "default":
        plt.style.use((pathlib.Path(__file__).parent / "matplotlibrc").resolve())


def save_figure(
    file_name: str = "figure",
    ext_list: Union[list, str, None] = None,
    white_background: bool = True,
) -> NoReturn:
    """Save matplotlib figure for all extensions in ext_list.

    Args:
        file_name: name of saved image file.
        ext_list: list of strings (or single string) denoting file type.
        white_background: set background of image to white if True.
    """

    # Default ext_list
    if ext_list is None:
        ext_list = ["pdf", "png"]

    # If ext_list is a single str
    if isinstance(ext_list, str):
        ext_list = [ext_list]

    # Set facecolor and edgecolor
    (fc, ec) = ("w", "w") if white_background else ("none", "none")

    # Save each type in ext_list
    for ext in ext_list:
        save_str = file_name + "." + ext
        plt.savefig(save_str, bbox_inches="tight", facecolor=fc, edgecolor=ec)
        print(f"Saved figure {save_str}")


def update_rc(key_str: str, value: Any) -> NoReturn:
    """Update matplotlibrc parameters.

    Args:
        key_str: string for a matplotlibrc parameter.
        value: associated value to set the matplotlibrc parameter.
    """
    plt.rcParams.update({key_str: value})



"""
Tests for data.
"""

from uncertainty_toolbox.data import (
    synthetic_arange_random,
    synthetic_sine_heteroscedastic,
)


def test_synthetic_arange_random_n_data():
    """Test if correct data quantity is generated by synthetic_arange_random."""
    n_list = [10, 20]
    for n in n_list:
        y_pred, y_std, y_true, x = synthetic_arange_random(n)
        assert len(y_pred) == n
        assert len(y_std) == n
        assert len(y_true) == n
        assert len(x) == n


def test_synthetic_sine_heteroscedastic_n_data():
    """Test if correct data quantity is generated by synthetic_sine_heteroscedastic."""
    n_list = [10, 20]
    for n in n_list:
        y_pred, y_std, y_true, x = synthetic_sine_heteroscedastic(n)
        assert len(y_pred) == n
        assert len(y_std) == n
        assert len(y_true) == n
        assert len(x) == n



"""
Metrics for assessing the quality of predictive uncertainty quantification.
"""
from typing import Any, Dict

import numpy as np

from uncertainty_toolbox.metrics_accuracy import prediction_error_metrics
from uncertainty_toolbox.metrics_calibration import (
    root_mean_squared_calibration_error,
    mean_absolute_calibration_error,
    miscalibration_area,
    adversarial_group_calibration,
    sharpness,
)
from uncertainty_toolbox.metrics_scoring_rule import (
    nll_gaussian,
    crps_gaussian,
    check_score,
    interval_score,
)


METRIC_NAMES = {
    "mae": "MAE",
    "rmse": "RMSE",
    "mdae": "MDAE",
    "marpd": "MARPD",
    "r2": "R2",
    "corr": "Correlation",
    "rms_cal": "Root-mean-squared Calibration Error",
    "ma_cal": "Mean-absolute Calibration Error",
    "miscal_area": "Miscalibration Area",
    "sharp": "Sharpness",
    "nll": "Negative-log-likelihood",
    "crps": "CRPS",
    "check": "Check Score",
    "interval": "Interval Score",
    "rms_adv_group_cal": ("Root-mean-squared Adversarial Group " "Calibration Error"),
    "ma_adv_group_cal": "Mean-absolute Adversarial Group Calibration Error",
}


def get_all_accuracy_metrics(
    y_pred: np.ndarray,
    y_true: np.ndarray,
    verbose: bool = True,
) -> Dict[str, float]:
    """Compute all accuracy metrics.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        verbose: Activate verbose mode.

    Returns:
        The evaluations for all accuracy related metrics.
    """
    if verbose:
        print(" (1/n) Calculating accuracy metrics")

    acc_metrics = prediction_error_metrics(y_pred, y_true)
    return acc_metrics


def get_all_average_calibration(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    num_bins: int,
    verbose: bool = True,
) -> Dict[str, float]:
    """Compute all metrics for average calibration.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of he predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        num_bins: The number of bins to use for discretization in some metrics.
        verbose: Activate verbose mode.

    Returns:
        The evaluations for all metrics relating to average calibration.
    """
    if verbose:
        print(" (2/n) Calculating average calibration metrics")

    cali_metrics = {}
    cali_metrics["rms_cal"] = root_mean_squared_calibration_error(
        y_pred, y_std, y_true, num_bins=num_bins
    )
    cali_metrics["ma_cal"] = mean_absolute_calibration_error(
        y_pred, y_std, y_true, num_bins=num_bins
    )
    cali_metrics["miscal_area"] = miscalibration_area(
        y_pred, y_std, y_true, num_bins=num_bins
    )

    return cali_metrics


def get_all_adversarial_group_calibration(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    num_bins: int,
    verbose: bool = True,
) -> Dict[str, Dict[str, np.ndarray]]:
    """Compute all metrics for adversarial group calibration.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of he predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        num_bins: The number of bins to use for discretization in some metrics.
        verbose: Activate verbose mode.

    Returns:
        The evaluations for all metrics relating to adversarial group calibration.
        Each inner dictionary contains the size of each group and the metrics
        computed for each group.
    """
    adv_group_cali_metrics = {}
    if verbose:
        print(" (3/n) Calculating adversarial group calibration metrics")
        print("  [1/2] for mean absolute calibration error")

    ma_adv_group_cali = adversarial_group_calibration(
        y_pred,
        y_std,
        y_true,
        cali_type="mean_abs",
        num_bins=num_bins,
        verbose=verbose,
    )
    ma_adv_group_size = ma_adv_group_cali.group_size
    ma_adv_group_cali_score_mean = ma_adv_group_cali.score_mean
    ma_adv_group_cali_score_stderr = ma_adv_group_cali.score_stderr
    adv_group_cali_metrics["ma_adv_group_cal"] = {
        "group_sizes": ma_adv_group_size,
        "adv_group_cali_mean": ma_adv_group_cali_score_mean,
        "adv_group_cali_stderr": ma_adv_group_cali_score_stderr,
    }

    if verbose:
        print("  [2/2] for root mean squared calibration error")

    rms_adv_group_cali = adversarial_group_calibration(
        y_pred,
        y_std,
        y_true,
        cali_type="root_mean_sq",
        num_bins=num_bins,
        verbose=verbose,
    )
    rms_adv_group_size = rms_adv_group_cali.group_size
    rms_adv_group_cali_score_mean = rms_adv_group_cali.score_mean
    rms_adv_group_cali_score_stderr = rms_adv_group_cali.score_stderr
    adv_group_cali_metrics["rms_adv_group_cal"] = {
        "group_sizes": rms_adv_group_size,
        "adv_group_cali_mean": rms_adv_group_cali_score_mean,
        "adv_group_cali_stderr": rms_adv_group_cali_score_stderr,
    }

    return adv_group_cali_metrics


def get_all_sharpness_metrics(
    y_std: np.ndarray,
    verbose: bool = True,
) -> Dict[str, float]:
    """Compute all sharpness metrics

    Args:
        y_std: 1D array of he predicted standard deviations for the held out dataset.
        verbose: Activate verbose mode.

    Returns:
        The evaluations for all sharpness metrics.
    """
    if verbose:
        print(" (4/n) Calculating sharpness metrics")

    sharp_metrics = {}
    sharp_metrics["sharp"] = sharpness(y_std)

    return sharp_metrics


def get_all_scoring_rule_metrics(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    resolution: int,
    scaled: bool,
    verbose: bool = True,
) -> Dict[str, float]:
    """Compute all scoring rule metrics

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of he predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        resolution: The number of quantiles to use for computation.
        scaled: Whether to scale the score by size of held out set.
        verbose: Activate verbose mode.

    Returns:
        The computed scoring rule metrics.
    """
    if verbose:
        print(" (n/n) Calculating proper scoring rule metrics")

    sr_metrics = {}
    sr_metrics["nll"] = nll_gaussian(y_pred, y_std, y_true, scaled=scaled)
    sr_metrics["crps"] = crps_gaussian(y_pred, y_std, y_true, scaled=scaled)
    sr_metrics["check"] = check_score(
        y_pred, y_std, y_true, scaled=scaled, resolution=resolution
    )
    sr_metrics["interval"] = interval_score(
        y_pred, y_std, y_true, scaled=scaled, resolution=resolution
    )

    return sr_metrics


def _print_adversarial_group_calibration(adv_group_metric_dic, print_group_num=3):
    for a_group_cali_type, a_group_cali_dic in adv_group_metric_dic.items():
        num_groups = a_group_cali_dic["group_sizes"].shape[0]
        print_idxs = [int(x) for x in np.linspace(1, num_groups - 1, print_group_num)]
        print("  {}".format(METRIC_NAMES[a_group_cali_type]))
        for idx in print_idxs:
            print(
                "     Group Size: {:.2f} -- Calibration Error: {:.3f}".format(
                    a_group_cali_dic["group_sizes"][idx],
                    a_group_cali_dic["adv_group_cali_mean"][idx],
                )
            )


def get_all_metrics(
    y_pred: np.ndarray,
    y_std: np.ndarray,
    y_true: np.ndarray,
    num_bins: int = 100,
    resolution: int = 99,
    scaled: bool = True,
    verbose: bool = True,
) -> Dict[str, Any]:
    """Compute all metrics.

    Args:
        y_pred: 1D array of the predicted means for the held out dataset.
        y_std: 1D array of he predicted standard deviations for the held out dataset.
        y_true: 1D array of the true labels in the held out dataset.
        num_bins: The number of bins to use for discretization in some metrics.
        resolution: The number of quantiles to use for computation.
        scaled: Whether to scale the score by size of held out set.
        verbose: Activate verbose mode.

    Returns:
        Dictionary containing all metrics.
    """
    # Accuracy
    accuracy_metrics = get_all_accuracy_metrics(y_pred, y_true, verbose)

    # Calibration
    calibration_metrics = get_all_average_calibration(
        y_pred, y_std, y_true, num_bins, verbose
    )

    # Adversarial Group Calibration
    adv_group_cali_metrics = get_all_adversarial_group_calibration(
        y_pred, y_std, y_true, num_bins, verbose
    )

    # Sharpness
    sharpness_metrics = get_all_sharpness_metrics(y_std, verbose)

    # Proper Scoring Rules
    scoring_rule_metrics = get_all_scoring_rule_metrics(
        y_pred, y_std, y_true, resolution, scaled, verbose
    )

    # Print all outputs
    if verbose:
        print("**Finished Calculating All Metrics**")
        print("\n")
        print(" Accuracy Metrics ".center(60, "="))
        for acc_metric, acc_val in accuracy_metrics.items():
            print("  {:<13} {:.3f}".format(METRIC_NAMES[acc_metric], acc_val))
        print(" Average Calibration Metrics ".center(60, "="))
        for cali_metric, cali_val in calibration_metrics.items():
            print("  {:<37} {:.3f}".format(METRIC_NAMES[cali_metric], cali_val))
        print(" Adversarial Group Calibration Metrics ".center(60, "="))
        _print_adversarial_group_calibration(adv_group_cali_metrics, print_group_num=3)
        print(" Sharpness Metrics ".center(60, "="))
        for sharp_metric, sharp_val in sharpness_metrics.items():
            print("  {:}   {:.3f}".format(METRIC_NAMES[sharp_metric], sharp_val))
        print(" Scoring Rule Metrics ".center(60, "="))
        for sr_metric, sr_val in scoring_rule_metrics.items():
            print("  {:<25} {:.3f}".format(METRIC_NAMES[sr_metric], sr_val))

    all_scores = {
        "accuracy": accuracy_metrics,
        "avg_calibration": calibration_metrics,
        "adv_group_calibration": adv_group_cali_metrics,
        "sharpness": sharpness_metrics,
        "scoring_rule": scoring_rule_metrics,
    }

    return all_scores



import uncertainty_toolbox as uct

# Load an example dataset of 100 predictions, uncertainties, and observations
predictions, predictions_std, y, x = uct.data.synthetic_sine_heteroscedastic(100)

# Compute all uncertainty metrics
metrics = uct.metrics.get_all_metrics(predictions, predictions_std, y)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements/requirements.txt', 'r') as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name="uncertainty_toolbox",
    version="0.1.1rc2",
    author="Willie Neiswanger, Youngseog Chung, Ian Char, Han Guo",
    author_email="willie.neiswanger@gmail.com",
    description=("A python toolbox for predictive uncertainty quantification,"
                 " calibration, metrics, and visualization."),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/uncertainty-toolbox/uncertainty-toolbox",
    packages=setuptools.find_packages(),
    package_data={"uncertainty_toolbox": ["matplotlibrc"]},
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



"""
Examples of code for visualizations.
"""

import numpy as np
import matplotlib.pyplot as plt
import uncertainty_toolbox as uct


# Set plot style
uct.viz.set_style()
uct.viz.update_rc("text.usetex", True)  # Set to True for system latex
uct.viz.update_rc("font.size", 14)  # Set font size
uct.viz.update_rc("xtick.labelsize", 14)  # Set font size for xaxis tick labels
uct.viz.update_rc("ytick.labelsize", 14)  # Set font size for yaxis tick labels

# Set random seed
np.random.seed(11)

# Generate synthetic predictive uncertainty results
n_obs = 650
f, std, y, x = uct.synthetic_sine_heteroscedastic(n_obs)

# Save figure (set to True to save)
savefig = True


def make_plots(pred_mean, pred_std, plot_save_str="row"):
    """Make set of plots."""

    ylims = [-3, 3]
    n_subset = 50

    fig, axs = plt.subplots(1, 3, figsize=(17, 8))

    # Make xy plot
    axs[0] = uct.plot_xy(
        pred_mean, pred_std, y, x, n_subset=300, ylims=ylims, xlims=[0, 15], ax=axs[0]
    )

    # Make ordered intervals plot
    axs[1] = uct.plot_intervals_ordered(
        pred_mean, pred_std, y, n_subset=n_subset, ylims=ylims, ax=axs[1]
    )

    # Make calibration plot
    axs[2] = uct.plot_calibration(pred_mean, pred_std, y, ax=axs[2])

    # Adjust subplots spacing
    fig.subplots_adjust(wspace=0.25)

    # Save figure
    if savefig:
        uct.viz.save_figure(plot_save_str, "svg", white_background=True)


# List of predictive means and standard deviations
pred_mean_list = [f]

pred_std_list = [
    std * 0.5,  # overconfident
    std * 2.0,  # underconfident
    std,  # correct
]

# Loop through, make plots, and compute metrics
idx_counter = 0
for i, pred_mean in enumerate(pred_mean_list):
    for j, pred_std in enumerate(pred_std_list):
        mace = uct.mean_absolute_calibration_error(pred_mean, pred_std, y)
        rmsce = uct.root_mean_squared_calibration_error(pred_mean, pred_std, y)
        ma = uct.miscalibration_area(pred_mean, pred_std, y)

        idx_counter += 1
        make_plots(pred_mean, pred_std, f"row_{idx_counter}")

        print(f"MACE: {mace}, RMSCE: {rmsce}, MA: {ma}")



"""
Tests for accuracy metrics.
"""

import pytest
import numpy as np

from uncertainty_toolbox.metrics_accuracy import prediction_error_metrics


@pytest.fixture
def get_test_set():
    y_pred = np.array([1, 2, 3])
    y_std = np.array([0.1, 0.5, 1])
    y_true = np.array([1.25, 2.2, 2.8])
    return y_pred, y_std, y_true


def test_prediction_error_metric_fields(get_test_set):
    """Test if prediction error metrics have correct fields."""
    y_pred, y_std, y_true = get_test_set
    met_dict = prediction_error_metrics(y_pred, y_true)
    met_keys = met_dict.keys()
    assert len(met_keys) == 6

    met_str_list = ["mae", "rmse", "mdae", "marpd", "r2", "corr"]
    bool_list = [s in met_keys for s in met_str_list]
    assert all(bool_list)


def test_prediction_error_metric_values(get_test_set):
    """Test if prediction error metrics have correct values."""
    y_pred, y_std, y_true = get_test_set
    met_dict = prediction_error_metrics(y_pred, y_true)
    print(met_dict)
    assert met_dict["mae"] > 0.21 and met_dict["mae"] < 0.22
    assert met_dict["rmse"] > 0.21 and met_dict["rmse"] < 0.22
    assert met_dict["mdae"] >= 0.20 and met_dict["mdae"] < 0.21
    assert met_dict["marpd"] > 12 and met_dict["marpd"] < 13
    assert met_dict["r2"] > 0.88 and met_dict["r2"] < 0.89
    assert met_dict["corr"] > 0.99 and met_dict["corr"] < 1.0
